local _tl_compat; if (tonumber((_VERSION or ''):match('[%d.]*$')) or 0) < 5.3 then local p, m = pcall(require, 'compat53.module'); if p then _tl_compat = m end end; local io = _tl_compat and _tl_compat.io or io; local os = _tl_compat and _tl_compat.os or os

local cli = require("cerulean.cli")

local code = cli.run(
arg,
cli.Stdio.new(
function(text) io.stdout:write(text) end,
function(text) io.stderr:write(text) end,
function() return io.stdin:read("*a") end))


os.exit(code)
